
class iRODSMeta(object):

    def __init__(self, name, value, units=None, avu_id=None, create_time=None, modify_time=None):
        self.avu_id = avu_id
        self.name = name
        self.value = value
        self.units = units
        self.create_time = create_time
        self.modify_time = modify_time

    def __eq__(self, other):
        return tuple(self) == tuple(other)

    def __iter__(self):
        yield self.name
        yield self.value
        if self.units: yield self.units

    def __repr__(self):
        return "<iRODSMeta {avu_id} {name} {value} {units}>".format(**vars(self))


class iRODSMetaCollection(object):
    
    def _reset_metadata(self):
        metas = self._model_cls._get_metadata()
        self._meta = [iRODSMeta(*result) for result in metas]
    
    def __init__(self, model_cls, path):
        self._model_cls = model_cls
        self._path = path
        self._reset_metadata()
        
        
    def get_all(self, key):
        """
        Returns a list of iRODSMeta associated with a given key
        """
        if not isinstance(key, str):
            raise TypeError
        return [m for m in self._meta if m.name == key]
    

    def get_one(self, key):
        """
        Returns the iRODSMeta defined for a key. If there are none,
        or if there are more than one defined, raises KeyError
        """
        values = self.get_all(key)
        if not values:
            raise KeyError
        if len(values) > 1:
            raise KeyError
        return values[0]


    def _get_meta(self, *args):
        if not len(args):
            raise ValueError(
                "Must specify an iRODSMeta object or key, value, units)")
        return args[0] if len(args) == 1 else iRODSMeta(*args)
    
    
    def items(self):
        """
        Returns a list of iRODSMeta
        """
        return self._meta


    def keys(self):
        """
        Return a list of keys. Duplicates preserved
        """
        return [m.name for m in self._meta] 
    
    
    def remove_all(self, **opts):
        ops = []
        for meta in self._meta:
            op = {
                "operation": "remove",
                "attribute": meta.name,
                "value": meta.value
            }
            if meta.units:
                op["units"] = meta.units
            ops.append(op)
        self._model_cls.manager.modify_metadata(self._path, ops)
        self._reset_metadata()
        
    
    def add(self, *args):
        """
        Add as iRODSMeta to a key
        """
        meta = self._get_meta(*args)
        op = {
            "operation": "add",
            "attribute": meta.name,
            "value": meta.value
        }
        if meta.units:
            op["units"] = meta.units
            
        self._model_cls.manager.modify_metadata(self._path, [op])
        self._reset_metadata()


    def remove(self, *args, **opts):
        """
        Removes an iRODSMeta
        """
        meta = self._get_meta(*args)
        op = {
            "operation": "remove",
            "attribute": meta.name,
            "value": meta.value
        }
        if meta.units:
            op["units"] = meta.units
        
        self._model_cls.manager.modify_metadata(self._path, [op])
        self._reset_metadata()